from unittest.mock import Mock

import pytest
from cashews.backends.interface import Backend
from cashews.utils import check_speed


@pytest.mark.asyncio
async def test_speed_test():
    backend = Mock(wraps=Backend())

    result = await check_speed(backend, iters=10)

    assert len(result) == 2
    assert "get" in result
    assert "set" in result
    assert backend.set.call_count == 10
    assert backend.get.call_count == 10



import dataclasses
from collections import namedtuple
from datetime import date, datetime, timedelta
from decimal import Decimal

import pytest
from cashews.serialize import PickleSerializerMixin, UnSecureDataError

pytestmark = pytest.mark.asyncio


@dataclasses.dataclass()
class TestDC:
    test: str
    _: int


NT = namedtuple("NT", ("test", "name"), defaults=[False, "test"])


class DummyCache:
    def __init__(self):
        self.store = {}

    async def set(self, key: str, value, *args, **kwargs) -> None:
        self.store[key] = value

    async def get(self, key: str, *args, **kwargs):
        return self.store.get(key, None)

    async def delete(self, *keys):
        for key in keys:
            del self.store[key]

    async def mget(self, *keys):
        return [self.store.get(key, None) for key in keys]


class Cache(PickleSerializerMixin, DummyCache):
    pass


@pytest.fixture(name="cache")
def _cache():
    cache = Cache()
    cache._hash_key = b"test"
    return cache


@pytest.mark.parametrize(
    "value",
    (
        "test",
        b"test",
        b"1_1",
        1,
        1.234,
        Decimal("1.001"),
        True,
        False,
        "",
        0,
        None,
        {"hi": True},
        TestDC(test="test", _=1),
    ),
)
async def test_serialize_simple_value(value, cache):
    await cache.set("key", value)
    assert await cache.get("key") == value
    assert await cache.mget("key") == (value,)


@pytest.mark.parametrize("value", (NT(name="test", test="lol"), NT()))
async def test_serialize_collections_value(value, cache):
    await cache.set("key", value)
    assert await cache.get("key") == value
    assert (await cache.get("key")).name == "test"


@pytest.mark.parametrize(
    "value", (datetime(year=2000, month=1, day=10, hour=10), timedelta(days=10), date(year=2020, month=12, day=31))
)
async def test_serialize_dates_value(value, cache):
    await cache.set("key", value)
    assert await cache.get("key") == value


@pytest.mark.parametrize(
    "value", (["test", "to"], (TestDC("hay", _=2),), {"test"}, [(1, 2), (3, 4)], {1, 2, 5}, [{"test": True}])
)
async def test_serialize_array_value(value, cache):
    await cache.set("key", value)
    assert list(await cache.get("key")) == list(value)
    assert type(await cache.get("key")) == type(value)


@pytest.mark.parametrize(
    "value", (["test", b"to"], (TestDC("hay", _=2), Decimal("10.1")), {"test", 1}, [], [{1: True}, Decimal("0.1")])
)
async def test_serialize_array_diff_value(value, cache):
    await cache.set("key", value)
    assert list(await cache.get("key")) == list(value)
    assert type(await cache.get("key")) == type(value)


async def test_unsecure_value(cache):
    cache.store["key"] = b"cos\nsystem\n(S'echo hello world'\ntR."
    with pytest.raises(UnSecureDataError):
        await cache.get("key")
    assert "key" not in cache.store

    cache.store["key"] = b"_cos\nsystem\n(S'echo hello world'\ntR."
    with pytest.raises(UnSecureDataError):
        await cache.get("key")
    assert "key" not in cache.store


async def test_no_value(cache):
    assert await cache.get("key") is None


async def test_replace_values(cache):
    await cache.set("key", "key")
    cache.store["replace"] = cache.store["key"]

    with pytest.raises(UnSecureDataError):
        await cache.get("replace")
    assert "replace" not in cache.store


async def test_pickle_error_value(cache):
    cache.store["key"] = cache.get_sign("key", b"no_pickle_data", b"md5") + b"_" + b"no_pickle_data"
    assert await cache.get("key") is None
    assert "key" not in cache.store


async def test_set_no_ser(cache):
    empty = object()
    await cache.set("key_e", empty)


async def test_no_import_dc(cache):
    @dataclasses.dataclass
    class TestNoImport:
        test: str

    with pytest.raises(AttributeError):
        await cache.set("key_i", TestNoImport(test="test"))


Schema = None


async def test_data_change(cache):
    global Schema

    @dataclasses.dataclass
    class Schema:
        test: str

    await cache.set("key", Schema("test"))

    @dataclasses.dataclass
    class Schema:
        name: str

    # Check without data change
    await cache.set("key2", Schema("test"))
    value = await cache.get("key2")
    assert value.name == "test"

    # IN key class of data have changed so it should be invalid
    value = await cache.get("key")
    assert value is None



import asyncio
import re
import time
import uuid
from collections import Counter, defaultdict
from statistics import mean
from string import Formatter

from .backends.interface import Backend
from .key import get_templates_for, template_to_pattern


async def _get_latency(func, **kwargs) -> float:
    start = time.perf_counter()
    await func(**kwargs)
    return time.perf_counter() - start


async def _get_average_latency(func, iters: int = 100, **kwargs):
    latency = []
    for _ in range(iters):
        latency.append(await _get_latency(func, **kwargs))
    return {"min": min(latency), "max": max(latency), "mean": mean(latency)}


async def _get_results(backend: Backend, key: str, iters=100):
    return await _get_average_latency(backend.get, key=key, iters=iters)


async def _set_results(backend: Backend, key: str, iters=100):
    return await _get_average_latency(backend.set, key=key, value="1", expire=10, iters=iters)


async def check_speed(backend: Backend, iters: int = 100):
    key = f"check:{str(uuid.uuid4())}"
    return {
        "set": await _set_results(backend, key, iters=iters),
        "get": await _get_results(backend, key, iters=iters),
    }


async def get_size_of(backend: Backend, func):
    coros = []
    templates = get_templates_for(func)
    for template in templates:
        coros.append(backend.get_size_match(template_to_pattern(template)))
    return zip(templates, await asyncio.gather(*coros))


class _ReFormatter(Formatter):
    def get_value(self, key, args, kwds):
        try:
            return kwds[key]
        except KeyError:
            return f"(P?<{key}>.*)"


class _NonKeyFormatter(Formatter):
    def get_value(self, key, args, kwds):
        try:
            return kwds[key]
        except KeyError:
            return f"{{{key}}}"


non_key_formatter = _NonKeyFormatter()


def _create_reader(template: str, max_events):
    pattern = re.compile(_ReFormatter().format(template).encode())
    get_counter = defaultdict(Counter)
    set_counter = defaultdict(Counter)

    def reader(cmd: bytes, key: bytes):
        counter = get_counter if cmd == b"get" else set_counter
        reader._max_events -= 1
        match = pattern.match(key)
        if match:
            for name, value in match.groupdict().items():
                counter[name][value] += 1
        if reader._max_events < 0:
            raise StopIteration()

    reader._max_events = max_events
    return reader, (get_counter, set_counter)


async def usage(backend: Backend, func, max_events=10, **values):
    templates = get_templates_for(func)
    if not templates:
        raise ValueError("Function have no associate cache key")
    coros = []
    counters = []
    for template in templates:
        pattern = template_to_pattern(template, **values)
        reader_template = non_key_formatter.format(template, **values)
        reader, counter = _create_reader(reader_template, max_events=max_events)
        coros.append(backend.listen(pattern, "get", "set", reader=reader))
        # await backend.listen(pattern, "get", "set", reader=reader)
        counters.append(counter)
    if coros:
        await asyncio.gather(*coros, return_exceptions=True)
        return counters



import asyncio
import logging
import time
from datetime import datetime, timedelta
from functools import wraps
from typing import Any, Callable, Optional

from ..backends.interface import Backend
from ..key import get_cache_key, get_cache_key_template, register_template
from ..typing import FuncArgsType
from .defaults import CacheDetect, _default_store_condition, context_cache_detect

__all__ = ("early",)
logger = logging.getLogger(__name__)


def early(
    backend: Backend,
    ttl: int,
    func_args: FuncArgsType = None,
    key: Optional[str] = None,
    store: Optional[Callable[[Any, Any, Any], bool]] = None,
    prefix: str = "early",
):
    """
    Cache strategy that try to solve Cache stampede problem (https://en.wikipedia.org/wiki/Cache_stampede),
    With hot cache recalculate a result in background near expiration time
    Warning Not good at cold cache

    :param backend: cache backend
    :param ttl: duration in seconds to store a result
    :param func_args: arguments that will be used in key
    :param key: custom cache key, may contain alias to args or kwargs passed to a call
    :param store: callable object that determines whether the result will be saved or not
    :param prefix: custom prefix for key, default 'early'
    """
    store = _default_store_condition if store is None else store

    def _decor(func):
        _key_template = key
        if key is None:
            _key_template = f"{prefix}:{get_cache_key_template(func, func_args=func_args, key=key)}:{ttl}"
        register_template(func, _key_template)

        @wraps(func)
        async def _wrap(*args, _from_cache: CacheDetect = context_cache_detect, **kwargs):
            _cache_key = get_cache_key(func, _key_template, args, kwargs, func_args)
            cached = await backend.get(_cache_key)
            if cached is not None:
                _from_cache.set(_cache_key, ttl=ttl)
                expire_at, delta, result = cached
                if expire_at <= datetime.utcnow() and await backend.set(
                    _cache_key + ":hit", "1", expire=delta.total_seconds(), exist=False
                ):
                    logger.info("Recalculate cache for %s (exp_at %s)", _cache_key, expire_at)
                    asyncio.create_task(_get_result_for_early(backend, func, args, kwargs, _cache_key, ttl, store))
                    # await asyncio.sleep(0)  # let loop switch to upadete cache
                return result
            return await _get_result_for_early(backend, func, args, kwargs, _cache_key, ttl, store)

        return _wrap

    return _decor


async def _get_result_for_early(backend: Backend, func, args, kwargs, key, ttl: int, condition):
    start = time.perf_counter()
    result = await func(*args, **kwargs)
    if condition(result, args, kwargs):
        delta = timedelta(seconds=max([ttl - (time.perf_counter() - start) * 3, 0]))
        await backend.set(key, [datetime.utcnow() + delta, delta, result], expire=ttl)
    return result



from contextvars import ContextVar


def _default_store_condition(result, args, kwargs) -> bool:
    return result is not None


class CacheDetect:
    def __init__(self):
        self._value = {}

    def set(self, key: str, **kwargs):
        self._value[key] = kwargs

    def get(self):
        return self._value


_var = ContextVar("cashews")
_var.set(None)


class _ContextCacheDetect:
    @staticmethod
    def start():
        _var.set(CacheDetect())

    @staticmethod
    def set(key: str, **kwargs):
        var = _var.get()
        if var is not None:
            var.set(key, **kwargs)

    @staticmethod
    def get():
        var = _var.get()
        if var is not None:
            return var.get()
        return var


context_cache_detect = _ContextCacheDetect()



import uuid
from contextlib import asynccontextmanager
from typing import Any, AsyncIterator, Optional, Tuple, Union


class LockedException(Exception):
    pass


class Backend:
    def __init__(self, *args, **kwargs):
        ...

    async def init(self):
        ...

    async def close(self):
        ...

    async def set(
        self, key: str, value: Any, expire: Union[None, float, int] = None, exist: Optional[bool] = None
    ) -> bool:
        ...

    async def get(self, key: str) -> Any:
        ...

    async def get_many(self, *keys: str) -> Tuple[Any]:
        ...

    async def keys_match(self, pattern: str):
        ...

    async def incr(self, key: str) -> int:
        ...

    async def delete(self, key: str):
        ...

    async def delete_match(self, pattern: str):
        ...

    async def expire(self, key: str, timeout: Union[float, int]):
        ...

    async def get_expire(self, key: str) -> int:
        ...

    async def get_size(self, key: str) -> int:
        """
        Return size in bites that allocated by a value for given key
        """
        ...

    async def get_size_match(self, pattern: str) -> int:
        size = 0
        async for key in self.keys_match(pattern):
            size += await self.get_size(key)
        return size

    async def listen(self, pattern: str, *cmds, reader=None):
        ...

    async def ping(self, message: Optional[str] = None) -> str:
        ...

    async def clear(self):
        ...

    async def set_lock(self, key: str, value: Any, expire: Union[float, int]) -> bool:
        ...

    async def is_locked(self, key: str, wait: Union[None, int, float] = None, step: Union[int, float] = 0.1) -> bool:
        ...

    async def unlock(self, key, value) -> bool:
        ...

    @asynccontextmanager
    async def lock(self, key, expire):
        identifier = str(uuid.uuid4())
        lock = await self.set_lock(key, identifier, expire=expire)
        if not lock and await self.ping(b"TEST") == b"TEST":
            raise LockedException(f"Key {key} already locked")
        try:
            yield
        finally:
            await self.unlock(key, identifier)


class ProxyBackend(Backend):
    def __init__(self, target=None):
        self._target = target
        super().__init__()

    def set(self, key: str, value: Any, expire: Union[None, float, int] = None, exist: Optional[bool] = None) -> bool:
        return self._target.set(key, value, expire=expire, exist=exist)

    def get(self, key: str) -> Any:
        return self._target.get(key)

    def get_many(self, *keys: str) -> Tuple[Any]:
        return self._target.get_many(keys)

    def incr(self, key: str) -> int:
        return self._target.incr(key)

    def delete(self, key: str):
        return self._target.delete(key)

    def delete_match(self, pattern: str):
        return self._target.delete_match(pattern)

    def expire(self, key: str, timeout: Union[int, float]):
        return self._target.expire(key, timeout)

    def get_expire(self, key: str) -> int:
        return self._target.get_expire(key)

    def ping(self, message: Optional[str] = None) -> str:
        if message is not None:
            return self._target.ping(message)
        return self._target.ping()

    def clear(self):
        return self._target.clear()

    def close(self):
        return self._target.close()

    def set_lock(self, key: str, value: Any, expire: Union[float, int]) -> bool:
        return self._target.set_lock(key, value, expire)

    def is_locked(self, key: str, wait: Union[int, float, None] = None, step: Union[int, float] = 0.1) -> bool:
        return self._target.is_locked(key, wait=wait, step=step)

    def unlock(self, key: str, value: str) -> bool:
        return self._target.unlock(key, value)

    def keys_match(self, pattern: str):
        return self._target.keys_match(pattern)

    def get_size(self, key):
        return self._target.get_size(key)

    def listen(self, pattern: str, *cmds, reader=None):
        return self._target.listen(pattern, *cmds, reader=reader)



import asyncio
from unittest.mock import Mock

import pytest
from cashews.backends.memory import Backend, Memory
from cashews.cache_utils.circuit_braker import CircuitBreakerOpen, circuit_breaker
from cashews.cache_utils.defaults import CacheDetect, context_cache_detect
from cashews.cache_utils.early import early as early_cache
from cashews.cache_utils.fail import fail
from cashews.cache_utils.locked import locked as lock_cache
from cashews.cache_utils.rate import PerfDegradationException, hit, perf
from cashews.cache_utils.simple import cache

pytestmark = pytest.mark.asyncio
EXPIRE = 0.01


class CustomError(Exception):
    pass


@pytest.fixture()
async def backend():
    _cache = Memory()
    await _cache.init()
    return _cache


async def test_fail_cache_simple(backend):
    @fail(backend, ttl=EXPIRE, exceptions=CustomError, func_args=())
    async def func(fail=False):
        if fail:
            raise CustomError()
        return b"ok"

    assert await func() == b"ok"
    await asyncio.sleep(0.001)
    assert await func(fail=True) == b"ok"

    await asyncio.sleep(EXPIRE)
    with pytest.raises(CustomError):
        await func(fail=True)


async def test_circuit_breaker_simple(backend):
    @circuit_breaker(backend, ttl=EXPIRE * 10, errors_rate=5, period=1, func_args=())
    async def func(fail=False):
        if fail:
            raise CustomError()
        return b"ok"

    for _ in range(9):
        assert await func() == b"ok"

    with pytest.raises(CustomError):
        await func(fail=True)
    await asyncio.sleep(0)

    with pytest.raises(CircuitBreakerOpen):
        await func(fail=True)

    with pytest.raises(CircuitBreakerOpen):
        await func(fail=False)


async def test_cache_simple(backend):
    @cache(backend, ttl=EXPIRE, key="key")
    async def func(resp=b"ok"):
        return resp

    assert await func() == b"ok"

    await asyncio.sleep(0)
    assert await func(b"notok") == b"ok"

    await asyncio.sleep(EXPIRE)
    assert await func(b"notok") == b"notok"


async def test_cache_simple_key(backend):
    @cache(backend, ttl=EXPIRE, func_args=("some",))
    async def func(resp=b"ok", some="err"):
        return resp

    @cache(backend, ttl=1, func_args=("some",))
    async def func2(resp=b"ok", some="err"):
        return resp

    assert await func() == b"ok"
    assert await func(b"notok", some="test") == b"notok"

    await asyncio.sleep(0)
    assert await func(b"notok") == b"ok"

    assert await func2(b"notok") == b"notok"


async def test_cache_simple_cond(backend):
    mock = Mock()

    @cache(backend, ttl=EXPIRE, key="key", store=lambda x, *args, **kwargs: x == b"hit")
    async def func(resp=b"ok"):
        mock()
        return resp

    await func()

    assert mock.call_count == 1

    await func(b"notok")
    assert mock.call_count == 2

    await func(b"hit")
    assert mock.call_count == 3
    await func(b"hit")
    assert mock.call_count == 3


async def test_early_cache_simple(backend):
    @early_cache(backend, ttl=EXPIRE, key="key")
    async def func(resp=b"ok"):
        return resp

    assert await func() == b"ok"

    await asyncio.sleep(0)
    assert await func(b"notok") == b"ok"

    await asyncio.sleep(EXPIRE)
    assert await func(b"notok") == b"notok"


async def test_early_cache_parallel(backend):

    mock = Mock()

    @early_cache(backend, ttl=0.1, key="key")
    async def func(resp=b"ok"):
        await asyncio.sleep(0.01)
        mock(resp)
        return resp

    assert await func() == b"ok"

    assert mock.call_count == 1

    for _ in range(4):  # 0.01 (first) + 4 * 0.01 = 0.06 + 0.01(executing) 0.8 will execute
        await asyncio.sleep(0.01)
        await asyncio.gather(*[func() for _ in range(10)])

    assert mock.call_count == 1

    for _ in range(60):  # 0.01 (hit) + 60 * 0.001  = 0.07 - next hit
        await asyncio.sleep(0.001)
        await asyncio.gather(*[func() for _ in range(10)])

    assert mock.call_count in [3, 4]


async def test_lock_cache_parallel(backend):
    mock = Mock()

    @lock_cache(backend, key="key", step=0.01)
    async def func(resp=b"ok"):
        await asyncio.sleep(0.01)
        mock(resp)
        return resp

    for _ in range(4):
        await asyncio.gather(*[func() for _ in range(10)], return_exceptions=True)

    assert mock.call_count == 4


async def test_lock_cache_broken_backend():
    backend = Mock(wraps=Backend())
    mock = Mock()

    @lock_cache(backend, key="key", step=0.01)
    async def func(resp=b"ok"):
        await asyncio.sleep(0.01)
        mock(resp)
        return resp

    for _ in range(4):
        await asyncio.gather(*[func() for _ in range(10)])

    assert mock.call_count == 40


async def test_hit_cache(backend):
    mock = Mock()

    @hit(backend, ttl=10, cache_hits=10, key="test")
    async def func(resp=b"ok"):
        mock(resp)
        return resp

    await func()  # cache
    await asyncio.gather(*[func() for _ in range(10)])  # get 10 hits
    assert mock.call_count == 1
    await func()  # cache
    assert mock.call_count == 2

    await asyncio.gather(*[func() for _ in range(10)])
    assert mock.call_count == 2


async def test_hit_cache_early(backend):
    mock = Mock()

    @hit(backend, ttl=10, cache_hits=1, key="test", update_before=0)
    async def func(resp=b"ok"):
        mock(resp)
        return resp

    assert await func(b"1") == b"1"  # cache
    assert mock.call_count == 1

    assert await func(b"2") == b"1"  # cache
    assert mock.call_count == 1

    await asyncio.sleep(0)
    assert await func(b"3") == b"2"  # cache
    assert mock.call_count == 2


async def test_perf_cache(backend):
    mock = Mock()

    @perf(backend, key="test", ttl=0.1)
    async def func(s=0.01):
        await asyncio.sleep(s)
        mock()
        return "res"

    await asyncio.gather(*[func() for _ in range(10)])
    assert mock.call_count == 10
    await func(0.04)
    assert mock.call_count == 11
    with pytest.raises(PerfDegradationException):
        await func(0.001)  # long
        assert mock.call_count == 11

    # prev was slow so no hits
    with pytest.raises(PerfDegradationException):
        await asyncio.gather(*[func() for _ in range(1000)])

    assert mock.call_count == 11
    await asyncio.sleep(0.07)

    await func(0.009)
    assert mock.call_count == 12


async def test_cache_detect_simple(backend):
    @cache(backend, ttl=EXPIRE, key="key")
    async def func(resp=b"ok"):
        return resp

    cache_detect = CacheDetect()
    assert await func(_from_cache=cache_detect) == b"ok"
    assert cache_detect.get() == {}

    await asyncio.sleep(0)
    assert await func(b"notok", _from_cache=cache_detect) == b"ok"
    assert len(cache_detect.get()) == 1
    assert list(cache_detect.get().keys()) == [
        "key",
    ]


async def test_context_cache_detect_simple(backend):
    @cache(backend, ttl=EXPIRE, key="key")
    async def func(resp=b"ok"):
        return resp

    context_cache_detect.start()
    assert await func() == b"ok"
    assert context_cache_detect.get() == {}

    await asyncio.sleep(0)
    assert await func(b"notok") == b"ok"
    assert len(context_cache_detect.get()) == 1
    assert list(context_cache_detect.get().keys()) == [
        "key",
    ]

    await asyncio.sleep(EXPIRE)
    assert await func(b"notok") == b"notok"
    assert len(context_cache_detect.get()) == 1


async def test_context_cache_detect_deep(backend):
    @cache(backend, ttl=EXPIRE, key="key1")
    async def func1():
        return 1

    @cache(backend, ttl=EXPIRE, key="key2")
    async def func2():
        return 2

    async def func():
        return await asyncio.gather(func1(), func2())

    context_cache_detect.start()
    await func()
    assert context_cache_detect.get() == {}

    await asyncio.sleep(0)
    await func()

    assert len(context_cache_detect.get()) == 2
    assert "key1" in context_cache_detect.get()
    assert "key2" in context_cache_detect.get()


async def test_context_cache_detect_context(backend):
    @cache(backend, ttl=EXPIRE, key="key1")
    async def func1():
        return 1

    @cache(backend, ttl=EXPIRE, key="key2")
    async def func2():
        return 2

    async def func(t_func):
        context_cache_detect.start()
        assert len(context_cache_detect.get()) == 0
        await t_func()
        assert len(context_cache_detect.get()) == 1

    await backend.set("key1", "test")
    await backend.set("key2", "test")
    assert await func1() == "test"
    assert await func2() == "test"

    context_cache_detect.start()
    await asyncio.gather(func(func1), func(func2), func1())
    assert len(context_cache_detect.get()) == 1  # ! It should be 3 but for now  it is ok, have no user case



from functools import wraps
from typing import Optional, Tuple, Type, Union

from ..backends.interface import Backend
from ..key import get_cache_key, get_cache_key_template, register_template
from ..typing import FuncArgsType
from .defaults import CacheDetect, context_cache_detect

__all__ = ("fail",)


def fail(
    backend: Backend,
    ttl: int,
    exceptions: Union[Type[Exception], Tuple[Type[Exception]]] = Exception,
    key: Optional[str] = None,
    func_args: FuncArgsType = None,
    prefix: str = "fail",
):
    """
    Return cache result (at list 1 call of function call should be succeed) if call raised one of given exception,
    :param backend: cache backend
    :param ttl: duration in seconds to store a result
    :param func_args: arguments that will be used in key
    :param exceptions: exceptions at which returned cache result
    :param key: custom cache key, may contain alias to args or kwargs passed to a call
    :param prefix: custom prefix for key, default "fail"
    """

    def _decor(func):
        _key_template = key
        if key is None:
            _key_template = f"{prefix}:{get_cache_key_template(func, func_args=func_args, key=key)}:{ttl}"
        register_template(func, _key_template)

        @wraps(func)
        async def _wrap(*args, _from_cache: CacheDetect = context_cache_detect, **kwargs):
            _cache_key = get_cache_key(func, _key_template, args, kwargs, func_args)
            try:
                result = await func(*args, **kwargs)
            except exceptions as exc:
                cached = await backend.get(_cache_key)
                if cached is not None:
                    _from_cache.set(_cache_key, ttl=ttl, exc=exc)
                    return cached
                raise exc
            else:
                await backend.set(_cache_key, result, expire=ttl)
                return result

        return _wrap

    return _decor



from datetime import datetime, timedelta

from .key import get_call_values


def at(hour=23, minute=59, second=59, minimum: timedelta = timedelta(minutes=1), **kwargs):
    def _ttl():
        now = datetime.utcnow()
        expire_at = now.replace(hour=hour, minute=minute, second=second, **kwargs)
        delta = expire_at - now
        if delta < minimum:
            return minimum
        return delta

    return _ttl


async def _is_disable_middleware(call, *args, backend=None, cmd=None, **kwargs):
    if backend.is_disable(cmd, "cmds"):
        return None
    return await call(*args, **kwargs)


async def _auto_init(call, *args, backend=None, cmd=None, **kwargs):
    if not backend.is_init:
        await backend._init()
    return await call(*args, **kwargs)


def add_prefix(prefix: str):
    async def _middleware(call, *args, backend=None, cmd=None, **kwargs):
        if cmd.lower() == "get_many":
            return await call(*[prefix + key for key in args])
        call_values = get_call_values(call, args, kwargs, func_args=None)
        as_key = "key"
        if cmd == "delete_match":
            as_key = "pattern"
        key = call_values.get(as_key)
        if key:
            call_values[as_key] = prefix + key
            return await call(**call_values)
        return await call(*args, **kwargs)

    return _middleware



import asyncio
import logging
import time
from collections import deque
from functools import wraps
from statistics import mean
from typing import Any, Callable, Iterable, Optional

from ..backends.interface import Backend
from ..key import get_cache_key, get_cache_key_template, register_template
from ..typing import FuncArgsType
from .defaults import CacheDetect, _default_store_condition, context_cache_detect

__all__ = ("hit", "perf", "rate_limit", "RateLimitException", "PerfDegradationException")


logger = logging.getLogger(__name__)


def hit(
    backend: Backend,
    ttl: int,
    cache_hits: int,
    update_before: Optional[int] = None,
    func_args: FuncArgsType = None,
    key: Optional[str] = None,
    store: Optional[Callable[[Any], bool]] = None,
    prefix: str = "hit",
):
    """
    Cache call results and drop cache after given numbers of call 'cache_hits'
    :param backend: cache backend
    :param ttl: duration in seconds to store a result
    :param cache_hits: number of cache hits till cache will dropped
    :param update_before: number of cache hits before cache will update
    :param func_args: arguments that will be used in key
    :param key: custom cache key, may contain alias to args or kwargs passed to a call
    :param disable: callable object that determines whether cache will use
    :param store: callable object that determines whether the result will be saved or not
    :param prefix: custom prefix for key, default 'hit'
    """
    store = _default_store_condition if store is None else store

    def _decor(func):
        _key_template = key
        if key is None:
            _key_template = f"{prefix}:{get_cache_key_template(func, func_args=func_args, key=key)}:{ttl}"
        register_template(func, _key_template)

        @wraps(func)
        async def _wrap(*args, _from_cache: CacheDetect = context_cache_detect, **kwargs):
            _cache_key = get_cache_key(func, _key_template, args, kwargs, func_args)
            result = await backend.get(_cache_key)
            hits = await backend.incr(_cache_key + ":counter")
            if result and hits <= cache_hits:
                _from_cache.set(_cache_key, ttl=ttl, cache_hits=cache_hits)
                if update_before is not None and cache_hits - hits <= update_before:
                    asyncio.create_task(_get_and_save(func, args, kwargs, backend, _cache_key, ttl, store))
                return result
            return await _get_and_save(func, args, kwargs, backend, _cache_key, ttl, store)

        return _wrap

    return _decor


async def _get_and_save(func, args, kwargs, backend, key, ttl, store):
    result = await func(*args, **kwargs)
    if store(result, args, kwargs):
        await backend.delete(key + ":counter")
        await backend.set(key, result, expire=ttl)

    return result


def _default_perf_condition(current: float, previous: Iterable[float]) -> bool:
    return mean(previous) * 2 < current


class PerfDegradationException(Exception):
    pass


def perf(
    backend: Backend,
    ttl: int,
    func_args: FuncArgsType = None,
    key: Optional[str] = None,
    trace_size: int = 10,
    perf_condition: Optional[Callable[[float, Iterable[float]], bool]] = None,
    prefix: str = "perf",
):
    """
    Trace time execution of target and throw exception if it downgrade to given condition
    :param backend: cache backend
    :param ttl: duration in seconds to lock
    :param func_args: arguments that will be used in key
    :param key: custom cache key, may contain alias to args or kwargs passed to a call
    :param trace_size: the number of calls that are involved
    :param perf_condition: callable object that determines whether the result will be cached,
           default if doubled mean value of time execution less then current
    :param prefix: custom prefix for key, default 'perf'

    """
    perf_condition = _default_perf_condition if perf_condition is None else perf_condition
    call_results = deque([], maxlen=trace_size)

    def _decor(func):
        _key_template = key
        if key is None:
            _key_template = f"{prefix}:{get_cache_key_template(func, func_args=func_args, key=key)}:{ttl}"

        @wraps(func)
        async def _wrap(*args, **kwargs):
            _cache_key = get_cache_key(func, _key_template, args, kwargs, func_args)
            if await backend.is_locked(_cache_key + ":lock"):
                raise PerfDegradationException()
            start = time.perf_counter()
            result = await func(*args, **kwargs)
            takes = time.perf_counter() - start
            if len(call_results) == trace_size and perf_condition(takes, call_results):
                await backend.set_lock(_cache_key + ":lock", value=takes, expire=ttl)
                call_results.clear()
            else:
                call_results.append(takes)
            return result

        return _wrap

    return _decor


class RateLimitException(Exception):
    pass


def _default_action(*args, **kwargs):
    raise RateLimitException()


def rate_limit(
    backend: Backend,
    limit: int,
    period: int,
    ttl: int = None,
    func_args: FuncArgsType = None,
    action: Optional[Callable] = None,
    prefix="rate_limit",
):  # pylint: disable=too-many-arguments
    """
    Rate limit for function call. Do not call function if rate limit is reached, and call given action

    :param backend: cache backend
    :param limit: number of calls
    :param period: Period
    :param ttl: time ban, default == period
    :param func_args: arguments that will be used in key
    :param action: call when rate limit reached, default raise RateLimitException
    :param prefix: custom prefix for key, default 'rate_limit'
    """
    action = _default_action if action is None else action

    def decorator(func):
        _key_template = f"{prefix}:{get_cache_key_template(func, func_args=func_args)}"
        register_template(func, _key_template)

        @wraps(func)
        async def wrapped_func(*args, **kwargs):
            _cache_key = get_cache_key(func, _key_template, args, kwargs, func_args)

            requests_count = await backend.incr(key=_cache_key)  # set 1 if not exists
            if requests_count and requests_count > limit:
                if ttl and requests_count == limit + 1:
                    await backend.expire(key=_cache_key, timeout=ttl)
                logger.info("Rate limit reach for %s", _cache_key)
                action(*args, **kwargs)

            if requests_count == 1:
                await backend.expire(key=_cache_key, timeout=period)

            return await func(*args, **kwargs)

        return wrapped_func

    return decorator



import asyncio
import sys
from unittest.mock import Mock

import pytest
from cashews.backends.interface import Backend, ProxyBackend
from cashews.backends.memory import Memory

pytestmark = pytest.mark.asyncio


@pytest.fixture(name="cache")
async def _cache():
    return Memory()


async def test_set_get(cache):
    await cache.set("key", b"value")
    assert await cache.get("key") == b"value"


async def test_set_get_many(cache):
    await cache.set("key", b"value")
    assert await cache.get_many("key", "no_exists") == (b"value", None)


async def test_set_exist(cache):
    assert await cache.set("key", b"value")
    assert await cache.set("key", b"value", exist=True)
    assert not await cache.set("key2", b"value", exist=True)

    assert await cache.set("key2", b"value", exist=False)
    assert not await cache.set("key2", b"value", exist=False)


async def test_get_no_value(cache):
    assert await cache.get("key2") is None


async def test_incr(cache):
    assert await cache.incr("incr") == 1
    assert await cache.incr("incr") == 2
    assert await cache.get("incr") == 2


async def test_ping(cache):
    assert await cache.ping() == b"PONG"


async def test_expire(cache):
    await cache.set("key", b"value", expire=0.01)
    assert await cache.get("key") == b"value"
    await asyncio.sleep(0.01)
    assert await cache.get("key") is None


@pytest.mark.parametrize(
    ("method", "args", "defaults"),
    (
        ("get", ("key",), None),
        ("set", ("key", "value"), {"exist": None, "expire": None}),
        ("incr", ("key",), None),
        ("delete", ("key",), None),
        ("expire", ("key", 10), None),
        ("ping", (), None),
        ("clear", (), None),
        ("set_lock", ("key", "value", 10), None),
        ("unlock", ("key", "value"), None),
        ("is_locked", ("key",), {"wait": None, "step": 0.1}),
    ),
)
async def test_proxy_backend(method, args, defaults):
    target = Mock(wraps=Backend())
    backend = ProxyBackend(target=target)

    await getattr(backend, method)(*args)
    if defaults:
        getattr(target, method).assert_called_once_with(*args, **defaults)
    else:
        getattr(target, method).assert_called_once_with(*args)


async def test_delete_match(cache: Backend):
    await cache.set("pref:test:test", b"value")
    await cache.set("pref:value:test", b"value2")
    await cache.set("pref:-:test", b"-")
    await cache.set("pref:*:test", b"*")

    await cache.set("ppref:test:test", b"value3")
    await cache.set("pref:test:tests", b"value3")

    await cache.delete_match("pref:*:test")

    assert await cache.get("pref:test:test") is None
    assert await cache.get("pref:value:test") is None
    assert await cache.get("pref:-:test") is None
    assert await cache.get("pref:*:test") is None

    assert await cache.get("ppref:test:test") is not None
    assert await cache.get("pref:test:tests") is not None


async def test_get_size(cache: Backend):
    await cache.set("test", b"1")
    assert await cache.get_size("test") == sys.getsizeof(b"1")


async def test_get_size_pattern(cache: Backend):
    await cache.set("target:test", b"1")
    await cache.set("no:test", b"1")
    assert await cache.get_size_match("target:*") == sys.getsizeof(b"1")
    assert await cache.get_size_match("*") == sys.getsizeof(b"1") * 2


async def test_listen(cache: Backend):
    m = Mock()
    asyncio.create_task(cache.listen("test:*", "get", "delete", reader=m))
    await asyncio.sleep(0)
    await cache.set("test:10", b"1")
    await cache.set("key", b"1")

    await cache.get("key")
    await cache.get("test:10")

    await asyncio.sleep(0)

    m.assert_called_with("get", "test:10")



from .circuit_braker import CircuitBreakerOpen, circuit_breaker  # noqa
from .defaults import CacheDetect, context_cache_detect  # noqa
from .early import early  # noqa
from .fail import fail  # noqa
from .invalidate import invalidate  # noqa
from .locked import locked  # noqa
from .rate import PerfDegradationException, RateLimitException, hit, perf, rate_limit  # noqa
from .simple import cache  # noqa



import inspect
from datetime import timedelta
from itertools import chain
from string import Formatter
from typing import Any, Callable, Dict, Optional, Tuple, Union

from .typing import TTL, FuncArgsType


def ttl_to_seconds(ttl: Union[float, None, TTL]) -> Union[int, None, float]:
    timeout = ttl() if callable(ttl) else ttl
    return timeout.total_seconds() if isinstance(timeout, timedelta) else timeout


def get_cache_key(
    func: Callable,
    template: str = None,
    args: Tuple[Any] = (),
    kwargs: Optional[Dict] = None,
    func_args: FuncArgsType = None,
) -> str:
    """
    Get cache key name for function (:param func) called with args and kwargs
    if func_args is passed key build with parameters are included in func_args dict or tuple otherwise use all of them
    Used function module and name as prefix if key parameter not passed
    :param func: Target function
    :param args: call positional arguments
    :param kwargs: call keyword arguments
    :param func_args: arguments that will be used in key
    :return: cache key for call
    """
    kwargs = kwargs or {}
    key_values = get_call_values(func, args, kwargs, func_args)
    key_values = {k: str(v) if v is not None else "" for k, v in key_values.items()}
    _key_template = template or get_cache_key_template(func, func_args)
    return _key_template.format(**key_values).lower()


def get_func_params(func):
    signature = inspect.signature(func)
    for param_name, param in signature.parameters.items():
        if param.kind not in [inspect.Parameter.VAR_KEYWORD, inspect.Parameter.VAR_POSITIONAL]:
            yield param_name


def get_cache_key_template(func: Callable, func_args: FuncArgsType = None, key: Optional[str] = None) -> str:
    """
    Get cache key name for function (:param func) called with args and kwargs
    if func_args is passed key build with parameters are included in func_args dict or tuple otherwise use all of them
    Used function module and name as prefix if key parameter not passed
    :param func: Target function
    :param args: call positional arguments
    :param kwargs: call keyword arguments
    :param func_args: arguments that will be used in key
    :param key: template for key, may contain alias to args or kwargs passed to a call
    :return: cache key for call
    """
    if key:
        return key

    if func_args is not None:
        params = {param: "{" + param + "}" for param in func_args}
    else:
        params = {param_name: "{" + param_name + "}" for param_name in get_func_params(func)}
    return ":".join([func.__module__, func.__name__, *chain(*params.items())]).lower()


def get_call_values(func: Callable, args, kwargs, func_args: FuncArgsType) -> Dict:
    """
    Return dict with arguments and their values for function call with given positional and keywords arguments
    :param func: Target function
    :param args: call positional arguments
    :param kwargs: call keyword arguments
    :param func_args: arguments that will be included in results (transformation function for values if passed as dict)
    """
    key_values = {key: "" for key in func_args or []}
    for _key, _value in _get_call_values(func, args, kwargs).items():
        if func_args is None or _key in func_args:
            key_values[_key] = _value
            if isinstance(func_args, dict) and callable(func_args[_key]):
                key_values[_key] = func_args[_key](_value)
            if isinstance(key_values[_key], bytes):
                key_values[_key] = key_values[_key].decode()
    return key_values


def _get_call_values(func, args, kwargs):
    signature = inspect.signature(func).bind(*args, **kwargs)
    signature.apply_defaults()
    result = {}
    for _name, _value in signature.arguments.items():
        parameter: inspect.Parameter = signature.signature.parameters[_name]
        if parameter.kind == inspect.Parameter.VAR_KEYWORD:
            result.update(_value)
        elif parameter.kind == inspect.Parameter.VAR_POSITIONAL:
            continue
        else:
            result[_name] = _value
    return result


class _Star(Formatter):
    def get_value(self, key, args, kwds):
        try:
            return kwds[key]
        except KeyError:
            return "*"


def template_to_pattern(template, _formatter=_Star(), **values):
    return _formatter.format(template, **values).lower()


_REGIRSTER = {}


def register_template(func, template):
    _REGIRSTER.setdefault((func.__module__, func.__name__), set()).add(template)


def get_templates_for(func):
    return _REGIRSTER.get((func.__module__, func.__name__), set())



import hashlib
import hmac
import pickle


class UnSecureDataError(Exception):
    pass


class PickleSerializerMixin:
    _digestmods = {
        b"sha1": hashlib.sha1,
        b"md5": hashlib.md5,
        b"sha256": hashlib.sha256,
    }

    def __init__(self, *args, hash_key=None, digestmod=b"md5", **kwargs):
        super().__init__(*args, **kwargs)
        self._hash_key = hash_key
        self._digestmod = digestmod

    async def get(self, key):
        try:
            return self._process(await super().get(key), key)
        except UnSecureDataError:
            await super().delete(key)
            raise
        except (pickle.PickleError, AttributeError):
            await super().delete(key)
            return None

    def _process(self, value: bytes, key):
        if not value:
            return value
        if value.isdigit():
            return int(value)
        try:
            sign, value = value.split(b"_", 1)
        except ValueError:
            raise UnSecureDataError()
        sign, digestmod = self._get_digestmod(sign)
        expected_sign = self.get_sign(key, value, digestmod)
        if expected_sign != sign:
            raise UnSecureDataError()
        value = pickle.loads(value, fix_imports=False, encoding="bytes")
        repr(value)
        return value

    async def mget(self, *keys):
        values = []
        for key, value in zip(keys, await super().mget(*keys)):
            try:
                values.append(self._process(value, key))
            except UnSecureDataError:
                await super().delete(key)
                raise
            except (pickle.PickleError, AttributeError):
                await super().delete(key)
                values.append(None)
        return tuple(values)

    def get_sign(self, key: str, value: bytes, digestmod: bytes) -> bytes:
        if self._hash_key is None:
            return b""
        value = key.encode() + value
        return hmac.new(self._hash_key, value, self._digestmods[digestmod]).hexdigest().encode()

    def _get_digestmod(self, sign: bytes):
        digestmod = self._digestmod
        if b":" in sign:
            digestmod, sign = sign.split(b":")
        if digestmod not in self._digestmods:
            raise UnSecureDataError()
        return sign, digestmod

    async def set(self, key: str, value, *args, **kwargs):
        value = pickle.dumps(value, protocol=pickle.HIGHEST_PROTOCOL, fix_imports=False)
        sign = self.get_sign(key, value, self._digestmod)
        return await super().set(key, self._digestmod + b":" + sign + b"_" + value, *args, **kwargs)



import pytest
from cashews.backends.memory import Memory, MemoryInterval
from cashews.backends.redis import Redis
from cashews.wrapper import settings_url_parse


@pytest.mark.parametrize(
    ("url", "params"),
    (
        ("mem://", {"backend": Memory}),
        ("mem://?size=10&check_interval=0.01", {"backend": MemoryInterval, "size": 10, "check_interval": 0.01}),
        ("redis://localhost:9000/0", {"backend": Redis, "address": "redis://localhost:9000/0"}),
        ("redis://password@localhost:9000", {"backend": Redis, "address": "redis://password@localhost:9000"}),
        (
            "redis://localhost/0/?password=password",
            {"backend": Redis, "address": "redis://localhost/0/", "password": "password"},
        ),
        (
            "redis://localhost/0/?hash_key=secret&password=test&safe=1&minsize=3&create_connection_timeout=0.1",
            {
                "backend": Redis,
                "address": "redis://localhost/0/",
                "hash_key": "secret",
                "password": "test",
                "safe": True,
                "minsize": 3,
                "create_connection_timeout": 0.1,
            },
        ),
    ),
)
def test_url(url, params):
    assert settings_url_parse(url) == params



import asyncio
from functools import partial, wraps
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Type, Union
from urllib.parse import parse_qsl, urlparse

from . import cache_utils
from .backends.interface import Backend, ProxyBackend
from .backends.memory import Memory, MemoryInterval
from .helpers import _auto_init, _is_disable_middleware
from .key import ttl_to_seconds
from .typing import TTL, FuncArgsType

#  pylint: disable=too-many-public-methods


class Cache(ProxyBackend):
    def __init__(self):
        self.__init = False
        self.__address = None
        self._kwargs = {}
        self._disable: Union[bool, List] = False
        self.middlewares = (_is_disable_middleware, _auto_init)
        super().__init__()

    @property
    def is_init(self):
        return self.__init

    def is_disable(self, *cmds: str) -> bool:
        if isinstance(self._disable, bool):
            return self._disable
        for cmd in cmds:
            if cmd.lower() in [c.lower() for c in self._disable]:
                return True
        return False

    def is_enable(self, *cmds):
        return not self.is_disable(*cmds)

    def disable(self, *cmds: str):
        if self._disable is True:
            self._disable = ["cmds", "decorators"]
        if self._disable is False:
            self._disable = []
        self._disable.extend(cmds)

    def enable(self, *cmds: str):
        if self._disable is True:
            self._disable = ["cmds", "decorators"]
        if self._disable is False:
            self._disable = []
            return
        for cmd in cmds:
            if cmd in self._disable:
                self._disable.remove(cmd)

    @property
    def disable_info(self):
        return self._disable

    def setup(self, settings_url: str, middlewares: Tuple = (), **kwargs):
        self.middlewares = tuple(middlewares) + self.middlewares
        params = settings_url_parse(settings_url)
        params.update(kwargs)
        if "disable" in params:
            self._disable = params.pop("disable")
        else:
            self._disable = not params.pop("enable", True)
        if not isinstance(self._disable, bool):
            self._disable = list(self._disable)
        self._setup_backend(**params)

    def _setup_backend(self, backend: Type[Backend], **kwargs):
        if self._target:
            asyncio.create_task(self._target.close())
        self._target = backend(**kwargs)
        self.__init = False

    async def init(self, *args, **kwargs):
        self.setup(*args, **kwargs)
        await self._init()

    async def _init(self):
        if self.is_disable():
            return None
        if not self.__init:
            await self._target.init()
        self.__init = True

    def _with_middlewares(self, cmd: str, target):
        call = target
        for middleware in self.middlewares:
            call = partial(middleware, call, cmd=cmd, backend=self)
        return call

    def set(self, key: str, value: Any, expire: Union[float, None, TTL] = None, exist: Optional[bool] = None):
        return self._with_middlewares("set", self._target.set)(
            key=key, value=value, expire=ttl_to_seconds(expire), exist=exist
        )

    def get(self, key: str) -> Any:
        return self._with_middlewares("get", self._target.get)(key=key)

    def get_many(self, *keys: str):
        return self._with_middlewares("get_many", self._target.get_many)(*keys)

    def incr(self, key: str) -> int:
        return self._with_middlewares("incr", self._target.incr)(key=key)

    def delete(self, key: str):
        return self._with_middlewares("delete", self._target.delete)(key=key)

    def delete_match(self, pattern: str):
        return self._with_middlewares("delete_match", self._target.delete_match)(pattern=pattern)

    def expire(self, key: str, timeout: TTL):
        return self._with_middlewares("expire", self._target.expire)(key=key, timeout=ttl_to_seconds(timeout))

    def get_expire(self, key: str) -> int:
        return self._with_middlewares("get_expire", self._target.get_expire)(key=key)

    def set_lock(self, key: str, value: Any, expire: TTL) -> bool:
        return self._with_middlewares("lock", self._target.set_lock)(
            key=key, value=value, expire=ttl_to_seconds(expire)
        )

    def unlock(self, key: str, value: str) -> bool:
        return self._with_middlewares("unlock", self._target.unlock)(key=key, value=value)

    def listen(self, pattern: str, *cmds, reader=None):
        return self._with_middlewares("listen", self._target.listen)(pattern, *cmds, reader=reader)

    def ping(self, message: Optional[str] = None) -> str:
        return self._with_middlewares("ping", self._target.ping)(message=message)

    def clear(self):
        return self._with_middlewares("clear", self._target.clear)()

    def is_locked(self, key: str, wait: Union[float, None, TTL] = None, step: Union[int, float] = 0.1) -> bool:
        return self._with_middlewares("is_locked", self._target.is_locked)(
            key=key, wait=ttl_to_seconds(wait), step=step
        )

    def _wrap_on_enable(self, name, decorator):
        def _decorator(func):
            decorator(func)  # to register cache templates

            @wraps(func)
            async def _call(*args, **kwargs):
                if self.is_disable("decorators", name):
                    return await func(*args, **kwargs)
                return await decorator(func)(*args, **kwargs)

            return _call

        return _decorator

    # DecoratorS
    def rate_limit(
        self,
        limit: int,
        period: TTL,
        ttl: Optional[TTL] = None,
        func_args: FuncArgsType = None,
        action: Optional[Callable] = None,
        prefix="rate_limit",
    ):  # pylint: disable=too-many-arguments
        return self._wrap_on_enable(
            "rate_limit",
            cache_utils.rate_limit(
                self,
                limit=limit,
                period=ttl_to_seconds(period),
                ttl=ttl_to_seconds(ttl),
                func_args=func_args,
                action=action,
                prefix=prefix,
            ),
        )

    def __call__(
        self,
        ttl: TTL,
        func_args: FuncArgsType = None,
        key: Optional[str] = None,
        store: Optional[Callable[[Any], bool]] = None,
        prefix: str = "",
    ):
        return self._wrap_on_enable(
            prefix or "cache",
            cache_utils.cache(self, ttl=ttl_to_seconds(ttl), func_args=func_args, key=key, store=store, prefix=prefix),
        )

    cache = __call__

    def invalidate(self, target, args_map: Optional[Dict[str, str]] = None, defaults: Optional[Dict] = None):
        return self._wrap_on_enable(
            "cache", cache_utils.invalidate(self, target=target, args_map=args_map, defaults=defaults)
        )

    def fail(
        self,
        ttl: TTL,
        exceptions: Union[Type[Exception], Tuple[Type[Exception]]] = Exception,
        key: Optional[str] = None,
        func_args: FuncArgsType = None,
        prefix: str = "fail",
    ):
        return self._wrap_on_enable(
            prefix,
            cache_utils.fail(
                self, ttl=ttl_to_seconds(ttl), exceptions=exceptions, key=key, func_args=func_args, prefix=prefix
            ),
        )

    def circuit_breaker(
        self,
        errors_rate: int,
        period: TTL,
        ttl: TTL,
        exceptions: Union[Type[Exception], Tuple[Type[Exception]]] = Exception,
        key: Optional[str] = None,
        func_args: FuncArgsType = None,
        prefix: str = "circuit_breaker",
    ):

        return self._wrap_on_enable(
            prefix,
            cache_utils.circuit_breaker(
                self,
                errors_rate=errors_rate,
                period=ttl_to_seconds(period),
                ttl=ttl_to_seconds(ttl),
                exceptions=exceptions,
                key=key,
                func_args=func_args,
                prefix=prefix,
            ),
        )

    def early(
        self,
        ttl: TTL,
        func_args: FuncArgsType = None,
        key: Optional[str] = None,
        store: Optional[Callable[[Any], bool]] = None,
        prefix: str = "early",
    ):
        return self._wrap_on_enable(
            prefix,
            cache_utils.early(self, ttl=ttl_to_seconds(ttl), func_args=func_args, key=key, store=store, prefix=prefix),
        )

    def hit(
        self,
        ttl: TTL,
        cache_hits: int,
        update_before: int = 0,
        func_args: FuncArgsType = None,
        key: Optional[str] = None,
        store: Optional[Callable[[Any], bool]] = None,
        prefix: str = "hit",
    ):
        return self._wrap_on_enable(
            prefix,
            cache_utils.hit(
                self,
                ttl=ttl_to_seconds(ttl),
                cache_hits=cache_hits,
                update_before=update_before,
                func_args=func_args,
                key=key,
                store=store,
                prefix=prefix,
            ),
        )

    def dynamic(
        self,
        func_args: FuncArgsType = None,
        ttl: TTL = 60 * 60 * 24,
        key: Optional[str] = None,
        store: Optional[Callable[[Any], bool]] = None,
        prefix: str = "dynamic",
    ):
        return self._wrap_on_enable(
            prefix,
            cache_utils.hit(
                self, ttl=ttl, cache_hits=3, update_before=2, func_args=func_args, key=key, store=store, prefix=prefix,
            ),
        )

    def perf(
        self,
        ttl: TTL,
        func_args: FuncArgsType = None,
        key: Optional[str] = None,
        trace_size: int = 10,
        perf_condition: Optional[Callable[[float, Iterable[float]], bool]] = None,
        prefix: str = "perf",
    ):
        return self._wrap_on_enable(
            prefix,
            cache_utils.perf(
                self,
                ttl=ttl,
                func_args=func_args,
                key=key,
                trace_size=trace_size,
                perf_condition=perf_condition,
                prefix=prefix,
            ),
        )

    def locked(
        self,
        ttl: Optional[TTL] = None,
        func_args: FuncArgsType = None,
        key: Optional[str] = None,
        step: Union[int, float] = 0.1,
        prefix: str = "locked",
    ):
        return self._wrap_on_enable(
            prefix, cache_utils.locked(self, ttl=ttl, func_args=func_args, key=key, step=step, prefix=prefix)
        )


def _fix_params_types(params: Dict[str, str]) -> Dict[str, Union[str, int, bool, float]]:
    new_params = {}
    bool_keys = ("safe", "enable", "disable")
    true_values = (
        "1",
        "true",
    )
    for key, value in params.items():
        if key.lower() in bool_keys:
            value = value.lower() in true_values
        elif value.isdigit():
            value = int(value)
        else:
            try:
                value = float(value)
            except ValueError:
                pass
        new_params[key.lower()] = value
    return new_params


def settings_url_parse(url):
    params = {}
    parse_result = urlparse(url)
    params.update(dict(parse_qsl(parse_result.query)))
    params = _fix_params_types(params)
    if parse_result.scheme == "redis":
        from cashews.backends.redis import Redis

        params["backend"] = Redis
        params["address"] = parse_result._replace(query=None)._replace(fragment=None).geturl()
    elif parse_result.scheme == "mem":
        params["backend"] = Memory
        if params.get("check_interval"):
            params["backend"] = MemoryInterval
    return params



import asyncio
import logging
import socket
from typing import Any, Union

from aioredis import ConnectionsPool
from aioredis import Redis as Redis_
from aioredis import RedisError, util

from ..serialize import PickleSerializerMixin
from .interface import ProxyBackend

__all__ = "SafeRedis"
logger = logging.getLogger(__name__)
_UNLOCK = """
if redis.call("get",KEYS[1]) == ARGV[1] then
    return redis.call("del",KEYS[1])
else
    return 0
end
"""
_KEY_SPACE = "__keyspace@{db}__:{pattern}"
_KEY_SPACE_NOTIFY = "notify-keyspace-events"
_NOTIFY_CONFIG = "KA"


# pylint: disable=arguments-differ
# pylint: disable=abstract-method


class _SafeRedis(PickleSerializerMixin, Redis_):
    """
    High-level Redis interface.
    Don't throw exception anyway.
    """

    async def execute(self, command, *args, **kwargs):
        try:
            return await super().execute(command, *args, **kwargs)
        except (RedisError, socket.gaierror, OSError, asyncio.TimeoutError):
            logger.error("Redis down on command %s", command)
            return None


class _Redis(PickleSerializerMixin, Redis_):
    pass


class Redis(ProxyBackend):
    def __init__(self, address, hash_key, digestmod="md5", safe=False, **kwargs):
        self._safe = safe
        if isinstance(hash_key, str):
            hash_key = hash_key.encode()

        self._hash_key = hash_key
        if isinstance(digestmod, str):
            digestmod = digestmod.encode()
        self._digestmod = digestmod

        self._address = address
        self._kwargs = kwargs
        super().__init__()

    async def init(self):
        pool = create_pool(address=self._address, **self._kwargs)
        _target_class = _SafeRedis if self._safe else _Redis

        try:
            await pool._fill_free(override_min=False)
        except Exception:
            if not self._safe:
                pool.close()
                await pool.wait_closed()
                raise

        self._target = _target_class(pool, hash_key=self._hash_key, digestmod=self._digestmod)

    def get_many(self, *keys: str):
        return self._target.mget(keys[0], *keys[1:])

    def set(self, key: str, value: Any, expire: Union[None, float, int] = None, exist=None):
        if exist is True:
            exist = Redis_.SET_IF_EXIST
        elif exist is False:
            exist = Redis_.SET_IF_NOT_EXIST
        pexpire = None
        if isinstance(expire, float):
            pexpire = int(expire * 1000)
            expire = None
        return self._target.set(key, value, expire=expire, pexpire=pexpire, exist=exist)

    def get_expire(self, key: str) -> int:
        return self._target.ttl(key)

    def clear(self):
        return self._target.flushdb()

    def set_lock(self, key: str, value, expire):
        return self.set(key, value, expire=expire, exist=False)

    async def is_locked(self, key: str, wait=None, step=0.1):
        if wait is None:
            return await self._target.exists(key)
        while wait > 0.0:
            if not await self._target.exists(key):
                return False
            wait -= step
            await asyncio.sleep(step)
        return True

    def unlock(self, key, value):
        return self._target.eval(_UNLOCK, keys=[key], args=[value])

    def delete(self, key: str):
        return self._target.unlink(key)

    async def keys_match(self, pattern: str):
        cursor = b"0"
        while cursor:
            cursor, keys = await self._target.scan(cursor, match=pattern, count=100)
            for key in keys:
                yield key

    async def delete_match(self, pattern: str):
        keys = []
        async for key in self.keys_match(pattern):
            keys.append(key)
        await self._target.unlink(*keys)

    async def listen(self, pattern: str, *cmds, reader=None):
        old = await self._target.config_get(_KEY_SPACE_NOTIFY)
        if old != _NOTIFY_CONFIG:
            await self._target.config_set(_KEY_SPACE_NOTIFY, _NOTIFY_CONFIG)
        (channel,) = await self._target.psubscribe(_KEY_SPACE.format(pattern=pattern, db=self._target.db))
        try:
            async for _channel, cmd in channel.iter():
                if cmd.decode() not in cmds:
                    continue
                reader(cmd, _channel.split(b":", 1)[1])
        finally:
            await self._target.punsubscribe(_KEY_SPACE.format(pattern=pattern, db=self._target.db))
            # if old != _NOTIFY_CONFIG:
            #     await self._target.config_set(_KEY_SPACE_NOTIFY, old)

    async def get_size(self, key: str) -> int:
        return int(await self._target.execute("MEMORY", "USAGE", key))

    async def close(self):
        if self._target:
            self._target.close()


def create_pool(
    address,
    *,
    db=None,
    password=None,
    ssl=None,
    encoding=None,
    minsize=1,
    maxsize=10,
    parser=None,
    loop=None,
    create_connection_timeout=None,
    pool_cls=None,
    connection_cls=None
):
    if pool_cls:
        cls = pool_cls
    else:
        cls = ConnectionsPool
    if isinstance(address, str):
        address, options = util.parse_url(address)
        db = options.setdefault("db", db)
        password = options.setdefault("password", password)
        encoding = options.setdefault("encoding", encoding)
        create_connection_timeout = options.setdefault("timeout", create_connection_timeout)
        if "ssl" in options:
            assert options["ssl"] or (not options["ssl"] and not ssl), (
                "Conflicting ssl options are set",
                options["ssl"],
                ssl,
            )
            ssl = ssl or options["ssl"]

    return cls(
        address,
        db,
        password,
        encoding,
        minsize=minsize,
        maxsize=maxsize,
        ssl=ssl,
        parser=parser,
        create_connection_timeout=create_connection_timeout,
        connection_cls=connection_cls,
        loop=loop,
    )



from unittest.mock import Mock, patch

import pytest
from cashews.backends.interface import Backend
from cashews.helpers import add_prefix
from cashews.wrapper import Cache


@pytest.fixture(name="target")
def _target():
    return Mock(wraps=Backend())


@pytest.fixture(name="cache")
def __cache(target):
    _cache = Cache()
    _cache._disable = False
    _cache._target = target
    return _cache


def test_setup(cache):
    backend = Mock()
    cache._target = None
    with patch("cashews.wrapper.Memory", backend):
        cache.setup("mem://localhost?test=1")
    backend.assert_called_with(test=1)


@pytest.mark.asyncio
async def test_init_disable(cache):
    backend = Mock(wraps=Backend())
    with patch("cashews.wrapper.Memory", Mock(return_value=backend)):
        await cache.init("mem://localhost?disable=1")
    backend.init.assert_not_called()
    assert cache.is_disable()


@pytest.mark.asyncio
async def test_disable_cmd(cache):
    await cache.init("mem://localhost")
    cache.disable("incr")
    await cache.set("test", 10)
    await cache.incr("test")
    assert await cache.get("test") == 10

    cache.enable("incr")
    await cache.incr("test")
    assert await cache.get("test") == 11


@pytest.mark.asyncio
async def test_disable_decorators(cache: Cache, target):
    cache.disable("decorators")
    data = (i for i in range(10))

    @cache(ttl=1)
    @cache.fail(1)
    @cache.hit(ttl=1, cache_hits=1)
    @cache.perf(ttl=1)
    @cache.circuit_breaker(ttl=1, errors_rate=1, period=1)
    @cache.rate_limit(ttl=1, limit=1, period=1)
    @cache.early(ttl=1)
    @cache.dynamic()
    @cache.locked(ttl=1)
    async def func():
        return next(data)

    assert await func() == 0
    assert await func() == 1
    target.get.assert_not_called()
    target.set.assert_not_called()

    cache.enable("decorators")
    assert await func() == 2
    target.get.assert_called()
    target.set.assert_called()


@pytest.mark.asyncio
async def test_init(cache):
    backend = Mock(wraps=Backend())
    with patch("cashews.wrapper.Memory", Mock(return_value=backend)):
        await cache.init("mem://localhost")
    backend.init.assert_called_once()
    assert cache.is_enable()


@pytest.mark.asyncio
async def test_auto_init(cache, target):
    await cache.ping()
    target.init.assert_called_once()


@pytest.mark.asyncio
async def test_add_prefix(cache: Cache, target):
    cache.middlewares = (add_prefix("prefix!"),)

    await cache.get(key="key")
    target.get.assert_called_once_with(key="prefix!key")

    await cache.set(key="key", value="value")
    target.set.assert_called_once_with(
        key="prefix!key", value="value", exist=None, expire=None,
    )
    await cache.ping()
    target.ping.assert_called_once_with(message=None)


@pytest.mark.asyncio
async def test_add_prefix_get_many(cache: Cache, target):
    cache.middlewares = (add_prefix("prefix!"),)
    await cache.get_many("key")
    target.get_many.assert_called_once_with("prefix!key")


@pytest.mark.asyncio
async def test_add_prefix_delete_match(cache: Cache, target):
    cache.middlewares = (add_prefix("prefix!"),)
    await cache.delete_match("key")
    target.delete_match.assert_called_once_with(pattern="prefix!key")


@pytest.mark.asyncio
async def test_smoke_cmds(cache: Cache, target):
    await cache.set(key="key", value={"any": True}, expire=60, exist=None)
    target.set.assert_called_once_with(key="key", value={"any": True}, expire=60, exist=None)

    await cache.get("key")  # -> Any
    target.get.assert_called_once_with(key="key")

    await cache.get_many("key1", "key2")
    target.get_many.assert_called_once_with("key1", "key2")

    await cache.incr("key")  # -> int
    target.incr.assert_called_once_with(key="key")

    await cache.delete("key")
    target.delete.assert_called_once_with(key="key")

    await cache.expire(key="key", timeout=10)
    target.expire.assert_called_once_with(key="key", timeout=10)

    await cache.get_expire(key="key")  # -> int seconds to expire
    target.get_expire.assert_called_once()

    await cache.ping(message="test")  # -> bytes
    target.ping.assert_called_once_with(message="test")

    await cache.clear()
    target.clear.assert_called_once_with()

    await cache.is_locked("key", wait=60)  # -> bool
    target.is_locked.assert_called_once_with(key="key", wait=60, step=0.1)

    await cache.set_lock("key", "value", expire=60)  # -> bool
    target.set_lock.assert_called_once_with(key="key", value="value", expire=60)

    await cache.unlock("key", "value")  # -> bool
    target.unlock.assert_called_once_with(key="key", value="value")



from operator import attrgetter

import pytest
from cashews.key import get_cache_key, get_cache_key_template


def test_cache_func_key_dict():
    async def func(user):
        ...

    obj = type("user", (), {"name": b"test", "one": True})()

    assert get_cache_key(func, args=(obj,), func_args={"user": attrgetter("name")}) == "tests.test_key:func:user:test"

    obj.name = b"new"
    assert get_cache_key(func, args=(obj,), func_args={"user": attrgetter("name")}) == "tests.test_key:func:user:new"


@pytest.mark.parametrize(
    ("args", "kwargs", "func_args", "key"),
    (
        (
            ("a1", "a2", "a3"),
            {"kwarg1": "k1", "kwarg3": "k3"},
            ("arg1", "kwarg1", "kwarg3"),
            "tests.test_key:func:arg1:a1:kwarg1:k1:kwarg3:k3",
        ),
        # (
        #     ("a1", "a2", "a3"),
        #     {"kwarg1": "k1", "kwarg3": "k3"},
        #     None,
        #     "tests.test_key:func:arg1:a1:arg2:a2:kwarg1:k1:kwarg2:true:kwarg3:k3",
        # ),
        (("a1", "a2", "a3"), None, ("arg1", "kwarg1", "kwarg3"), "tests.test_key:func:arg1:a1:kwarg1::kwarg3:"),
        (
            ("a1",),
            {"arg2": 2, "kwarg1": "k1"},
            ("arg2", "kwarg1", "kwarg3"),
            "tests.test_key:func:arg2:2:kwarg1:k1:kwarg3:",
        ),
        (
            ("a1",),
            {"arg2": 2, "kwarg1": "k1", "kwarg3": "k3"},
            ("arg2", "kwarg1", "kwarg3"),
            "tests.test_key:func:arg2:2:kwarg1:k1:kwarg3:k3",
        ),
        (
            ("a1",),
            {"kwarg1": "k1", "arg2": 2},
            ("arg2", "kwarg1", "kwarg3"),
            "tests.test_key:func:arg2:2:kwarg1:k1:kwarg3:",
        ),
    ),
)
def test_cache_key_args_kwargs(args, kwargs, func_args, key):
    async def func(arg1, arg2, *args, kwarg1=None, kwarg2=b"true", **kwargs):
        ...

    assert get_cache_key(func, args=args, kwargs=kwargs, func_args=func_args) == key


@pytest.mark.parametrize(
    ("args", "func_args", "result"),
    ((("a", "k"), ("a", "k"), "tests.test_key:func:a:a:k:k"), (("a",), ("a", "k"), "tests.test_key:func:a:a"),),
)
def atest_cache_key_key(args, func_args, result):
    async def func(a="a1", k=None):
        ...

    assert get_cache_key(func, args=args, func_args=func_args) == result


async def func1(a):
    ...


async def func2(a, k=None, **kwargs):
    ...


@pytest.mark.parametrize(
    ("func", "func_args", "key", "template"),
    (
        (func1, None, None, "tests.test_key:func1:a:{a}"),
        (func2, None, None, "tests.test_key:func2:a:{a}:k:{k}"),
        (func2, ("a",), None, "tests.test_key:func2:a:{a}"),
        (func2, ("k",), None, "tests.test_key:func2:k:{k}"),
        (func2, ("k", "test"), None, "tests.test_key:func2:k:{k}:test:{test}"),
        (func2, {"k": ""}, None, "tests.test_key:func2:k:{k}"),
        (func2, ("k1",), None, "tests.test_key:func2:k1:{k1}"),
        (func2, None, "key", "key"),
    ),
)
def test_get_key_template(func, func_args, key, template):
    assert get_cache_key_template(func, func_args, key) == template



import asyncio
import datetime
import gc
import re
import sys
from typing import Any, Optional, Tuple, Union

from .interface import Backend

__all__ = ("Memory", "MemoryInterval")


class Memory(Backend):
    def __init__(self, size: int = 1000):
        self.store = {}
        self._chans = []
        self.size = size
        self._meta = {}
        self._lock = asyncio.Lock()
        super().__init__()

    async def clear(self):
        self.store = {}

    async def set(self, key: str, value: Any, expire: Union[None, float, int] = None, exist=None) -> bool:
        if exist is not None:
            if not (key in self.store) is exist:
                return False
        async with self._lock:
            self._set(key, value, expire)
        return True

    async def get(self, key: str) -> Any:
        return self._get(key)

    async def get_many(self, *keys: str) -> Tuple:
        return tuple([self._get(key) for key in keys])

    async def keys_match(self, pattern: str):
        pattern = pattern.replace("*", ".*")
        regexp = re.compile(pattern)
        for key in dict(self.store):
            if regexp.fullmatch(key):
                yield key

    async def incr(self, key: str):
        value = int(self._get(key) or 0) + 1
        self._set(key=key, value=value)
        return value

    async def delete(self, key: str):
        async with self._lock:
            self._delete(key)

    def _delete(self, key: str) -> bool:
        if key in self.store:
            self._notify("delete", key)
            del self.store[key]
            if key in self._meta:
                del self._meta[key]
            return True
        return False

    async def delete_match(self, pattern: str):
        async for key in self.keys_match(pattern):
            await self.delete(key)

    async def expire(self, key: str, timeout: float):
        value = self._get(key)
        if value is None:
            return
        self._set(key=key, value=value, expire=timeout)

    async def get_expire(self, key: str) -> int:
        return -1

    async def ping(self, message: Optional[str] = None):
        self._notify("ping", message)
        return b"PONG" if message is None else message

    def _set(self, key: str, value: Any, expire: Optional[float] = None):
        if len(self.store) > self.size:
            return
        self._notify("set", key)
        self.store[key] = value

        if expire is not None and expire > 0.0:
            loop = asyncio.get_event_loop()
            handler = self._meta.get(key)
            if handler:
                handler.cancel()
            self._meta[key] = loop.call_later(expire, self._delete, key)

    def _get(self, key: str) -> Optional[Any]:
        self._notify("get", key)
        return self.store.get(key, None)

    async def set_lock(self, key: str, value, expire):
        return await self.set(key, value, expire=expire, exist=False)

    async def is_locked(self, key: str, wait=None, step=0.1) -> bool:
        if wait is None:
            return key in self.store
        while wait > 0:
            if key not in self.store:
                return False
            wait -= step
            await asyncio.sleep(step)
        return key in self.store

    async def unlock(self, key, value) -> bool:
        return self._delete(key)

    async def get_size(self, key: str) -> int:
        if key in self.store:
            return _get_obj_size(self.store[key])
        return 0

    def _notify(self, cmd, key):
        for queue in self._chans:
            queue.put_nowait((cmd, key))

    async def listen(self, pattern: str, *cmds, reader=None):
        queue = asyncio.Queue(maxsize=0)
        pattern = pattern.replace("*", ".*")
        regexp = re.compile(pattern)
        self._chans.append(queue)
        try:
            while True:
                cmd, key = await queue.get()
                if cmd in cmds and regexp.fullmatch(key):
                    reader(cmd, key)
        finally:
            self._chans.remove(queue)


class MemoryInterval(Memory):
    def __init__(self, size: int = 1000, check_interval: float = 1.0):
        self._check_interval = check_interval
        super().__init__(size=size)

    async def init(self):
        asyncio.create_task(self._remove_expired())

    async def _remove_expired(self):
        while True:
            store = dict(self.store)
            for key in store:
                await self.get(key)
            await asyncio.sleep(self._check_interval)

    async def get(self, key: str) -> Optional[Any]:
        expiration = self._meta.get(key)
        if expiration and datetime.datetime.utcnow() > expiration:
            await self.delete(key)
            return None
        self._notify("get", key)
        return self.store.get(key, None)

    def _set(self, key: str, value: Any, expire: Optional[float] = None) -> None:
        self._notify("set", key)
        self.store[key] = value

        if expire is not None and expire > 0.0:
            self._meta[key] = datetime.datetime.utcnow() + datetime.timedelta(seconds=expire)

    async def get_expire(self, key: str) -> int:
        if key in self._meta:
            return abs(int((datetime.datetime.utcnow() - self._meta[key]).total_seconds()))
        return -1


def _get_obj_size(obj) -> int:
    marked = {id(obj)}
    obj_q = [obj]
    size = 0

    while obj_q:
        size += sum(map(sys.getsizeof, obj_q))

        # Lookup all the object referred to by the object in obj_q.
        # See: https://docs.python.org/3.7/library/gc.html#gc.get_referents
        all_refr = ((id(o), o) for o in gc.get_referents(*obj_q))

        # Filter object that are already marked.
        # Using dict notation will prevent repeated objects.
        new_refr = {o_id: o for o_id, o in all_refr if o_id not in marked and not isinstance(o, type)}

        # The new obj_q will be the ones that were not marked,
        # and we will update marked with their ids so we will
        # not traverse them again.
        obj_q = new_refr.values()
        marked.update(new_refr.keys())

    return size



from setuptools import setup, find_packages


with open("Readme.md", "rt", encoding="utf8") as f:
    readme = f.read()

PROJECT_NAME = "cashews"
VERSION = "0.15.0"

setup(
    name=PROJECT_NAME,
    version=VERSION,
    author="Dmitry Kryukov",
    url="https://github.com/Krukov/" + PROJECT_NAME,
    download_url="https://github.com/Krukov/" + PROJECT_NAME + "/tarball/" + VERSION,
    author_email="glebov.ru@gmail.com",
    description="cache tools with async power",
    keywords="cache aio async multicache aiocache",
    long_description=readme,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Framework :: AsyncIO",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "Operating System :: OS Independent",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development",
        "Typing :: Typed",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
    ],
    packages=find_packages(),
    install_requires=None,
    extras_require={
        "redis": ["aioredis>=1.0.0"],
        "dev": ["black", "codecov", "coverage", "flake8", "pytest", "isort", "pylint", "pytest-asyncio"],
    },
)



from .backends.interface import LockedException
from .cache_utils import (  # noqa
    CacheDetect,
    CircuitBreakerOpen,
    PerfDegradationException,
    RateLimitException,
    context_cache_detect,
)
from .helpers import add_prefix, at  # noqa
from .wrapper import Cache  # noqa

# pylint: disable=invalid-name
cache = Cache()
rate_limit = cache.rate_limit
fail = cache.fail
circuit_breaker = cache.circuit_breaker
early = cache.early
hit = cache.hit
dynamic = cache.dynamic
perf = cache.perf
locked = cache.locked
invalidate = cache.invalidate


mem = Cache()
mem.setup(
    "mem://?check_interval=1", size=1_000_000, middlewares=(add_prefix("mem:"),),
)  # 1_000_000 * 248(size of small dict) == 31 mb



from datetime import timedelta
from typing import Any, Callable, Dict, Optional, Tuple, Union

FuncArgsType = Optional[Union[Dict[str, Callable[[Any], str]], Tuple[str]]]
TTL = Union[int, timedelta, Callable[[], Union[int, timedelta]]]
